from __future__ import annotations

from typing import Literal
from urllib.parse import unquote

import pandas as pd
from pandas import DataFrame
from pandas._typing import FilePath, ReadCsvBuffer

from .config import settings


def quote(s: str) -> str:
    return s.translate(
        str.maketrans(
            {  # type: ignore[arg-type]
                ";": "%3B",
                "=": "%3D",
                "&": "%26",
                ",": "%2C",
                chr(0x7F): "%7F",
                **{chr(i): f"%{i:x}".upper() for i in range(0x1F)},
            }
        )
    )


GFF3_COLUMNS = [
    "seqid",
    "source",
    "type",
    "start",
    "end",
    "score",
    "strand",
    "phase",
    "attributes",
]
GFF3_ATTRIBUTES = (
    "ID",
    "Name",
    "Alias",
    "Parent",
    "Target",
    "Gap",
    "Derives_from",
    "Note",
    "Dbxref",
    "Ontology_term",
    "Is_circular",
)


class GFFDataFrame(DataFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for col in GFF3_COLUMNS:
            if col not in self.columns:
                raise AttributeError(f"Column '{col}' is required")

    def to_gff3(
        self,
        gff_file: FilePath | None = None,
        *,
        extra: Literal["ignore", "merge", "inplace"] = "ignore",
    ) -> str | None:
        df = self[GFF3_COLUMNS].copy()
        if extra != "ignore":
            extra_fields = [
                {k: v for k, v in e.items() if k not in GFF3_COLUMNS}
                for e in self.to_dict(orient="records")
            ]
            if extra == "merge":
                attributes_field = [
                    a | e
                    for a, e in zip(
                        self.attributes.to_dict(orient="records"), extra_fields
                    )
                ]
            else:
                attributes_field = extra_fields
            df["attributes"] = [
                ";".join(
                    f"{quote(k)}={quote(str(v))}"
                    for k, v in r.items()
                    if isinstance(k, str) and not pd.isna(v)
                )
                for r in attributes_field
            ]
        gff_feature = df.to_csv(sep="\t", na_rep=".", index=False, header=False)
        if gff_file is None:
            return f"""##gff-version 3\n{gff_feature}"""
        with open(gff_file, "w") as fh:
            fh.write("##gff-version 3\n")
            fh.write(gff_feature)
            return None

    @property
    def version(self) -> str:
        return "3"

    @property
    def attributes(self) -> DataFrame:
        df = pd.json_normalize(
            self["attributes"].apply(
                lambda attributes: dict(kv.split("=") for kv in attributes.split(";"))
            )  # type: ignore
        ).map(lambda v: unquote(v) if isinstance(v, str) else v)
        df.index = self.index
        order: dict[str, int] = dict((a, i) for i, a in enumerate(GFF3_ATTRIBUTES))
        columns = sorted(list(df.columns), key=lambda c: order.get(c, len(c)))
        return df[columns]  # pyright: ignore

    def attributes_to_columns(self) -> GFFDataFrame:
        attributes = self.attributes
        df = self.copy()
        for c in attributes.columns:
            if c not in GFF3_COLUMNS:
                df[c] = attributes[c]
        return df


def read_gff3(
    input_file: FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str], **kwargs
):
    """Read GFF3 files.

    Parameters
    ----------
    input_file : str | os.PathLike | ReadableBuffer
        support fsspec chain (available in pandas 3.0)
    **kwargs
        will pass to `pd.read_table`

    Returns
    -------
    GFFDataFrame

    Raises
    ------
    ValueError
        if provided 'comment' parameter
    """
    for param in ("comment", "na_values"):
        if param in kwargs:
            raise ValueError(f"Parameter '{param}' is not allowed")
    kwargs["comment"] = "#"
    kwargs["na_values"] = "."
    if "names" not in kwargs:
        kwargs["names"] = GFF3_COLUMNS
    if isinstance(input_file, str):
        *protocols, path = input_file.split("::")
        # https://github.com/pandas-dev/pandas/pull/60100
        if any([protocol.startswith("tar://") for protocol in protocols]):
            kwargs["compression"] = None
        if (
            settings.cache_http
            and path.startswith(("https://", "http://"))
            and len(protocols) == 0
            or "filecache" != protocols[-1]
        ):
            input_file = "::".join([*protocols, "filecache", path])
    return GFFDataFrame(pd.read_table(input_file, **kwargs))



from typing import Any

from Bio.Seq import Seq as _Seq
from pydantic import (
    GetCoreSchemaHandler,
)
from pydantic_core import CoreSchema, core_schema


class Seq(_Seq):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_plain_validator_function(
            _Seq,
            serialization=core_schema.plain_serializer_function_ser_schema(
                str, return_schema=core_schema.str_schema()
            ),
        )



from pydantic import BaseModel, TypeAdapter

from biov.types import Seq


class A(BaseModel):
    seq: Seq


def test_seq_validation_and_serialization():
    acgt = TypeAdapter(Seq).validate_python("ACGT")
    assert acgt == "ACGT"
    assert acgt.reverse_complement() == "ACGT"
    a = A.model_validate({"seq": "ACGT"})
    assert a.seq == acgt
    assert a.model_dump() == {"seq": "ACGT"}
    assert a.model_dump_json() == '{"seq":"ACGT"}'



from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest

from biov.gff import GFF3_COLUMNS, GFFDataFrame, read_gff3

# Test data
SAMPLE_GFF = """##gff-version 3
chr1\t.\tgene\t100\t200\t.\t+\t.\tID=gene1;Name=TestGene;Dbxref=NCBI:123
chr1\t.\texon\t150\t180\t.\t+\t.\tID=exon1;Parent=gene1
"""


@pytest.fixture
def sample_gff_df():
    data = {
        "seqid": ["chr1", "chr1"],
        "source": [None, None],
        "type": ["gene", "exon"],
        "start": [100, 150],
        "end": [200, 180],
        "score": [None, None],
        "strand": ["+", "+"],
        "phase": [None, None],
        "attributes": [
            "ID=gene1;Name=TestGene;Dbxref=NCBI:123",
            "ID=exon1;Parent=gene1",
        ],
    }
    return GFFDataFrame(data)


def test_read_gff3():
    df = read_gff3(StringIO(SAMPLE_GFF))
    assert isinstance(df, GFFDataFrame)
    assert list(df.columns) == GFF3_COLUMNS
    assert len(df) == 2


def test_gffdataframe_validation():
    # Should pass with all required columns
    valid_data = {col: [] for col in GFF3_COLUMNS}
    GFFDataFrame(valid_data)

    # Should fail if missing columns
    invalid_data = {col: [] for col in GFF3_COLUMNS[:-1]}
    with pytest.raises(AttributeError):
        GFFDataFrame(invalid_data)


def test_to_gff3(sample_gff_df: GFFDataFrame):
    with NamedTemporaryFile(mode="w", suffix=".gff") as f:
        sample_gff_df.to_gff3(f.name)
        content = Path(f.name).read_text()
        assert content.startswith("##gff-version 3")
        assert "chr1\t.\tgene\t100\t200" in content
    sample_gff_df["hello"] = "world"
    assert sample_gff_df.to_gff3() == SAMPLE_GFF
    assert (
        sample_gff_df.to_gff3(extra="merge")
        == """##gff-version 3
chr1\t.\tgene\t100\t200\t.\t+\t.\tID=gene1;Name=TestGene;Dbxref=NCBI:123;hello=world
chr1\t.\texon\t150\t180\t.\t+\t.\tID=exon1;Parent=gene1;hello=world
"""
    )
    assert (
        sample_gff_df.to_gff3(extra="inplace")
        == """##gff-version 3
chr1\t.\tgene\t100\t200\t.\t+\t.\thello=world
chr1\t.\texon\t150\t180\t.\t+\t.\thello=world
"""
    )


def test_attributes_to_columns(sample_gff_df: GFFDataFrame):
    df = sample_gff_df.attributes_to_columns()
    assert "ID" in df.columns
    assert "Name" in df.columns
    assert "Parent" in df.columns
    assert df.loc[0, "ID"] == "gene1"
    assert df.loc[0, "Name"] == "TestGene"
    assert df.loc[1, "Parent"] == "gene1"



import os
from pathlib import Path

import fsspec.config
from pydantic_settings import BaseSettings


class Settings(BaseSettings, env_prefix="BIOV_"):
    home: Path | None = None
    cache_http: bool = True


settings = Settings()

appname = __package__ or "biov"

if settings.home is not None:
    cache_storage = str(settings.home)
elif (xdg_cache_home := os.getenv("XDG_CACHE_HOME")) is not None:
    cache_storage = os.path.join(xdg_cache_home, appname)
else:
    from platformdirs import PlatformDirs

    cache_storage = PlatformDirs(appname=appname).user_cache_dir

fsspec.config.conf["filecache"] = {"cache_storage": cache_storage}



import pandas.io.common

# if pandas dependencies upgrade to 3.0, this patch could be removed
if not pandas.io.common.is_fsspec_url("filecache::s3://some-bucket/some-file"):  # type: ignore
    import re

    # https://github.com/pandas-dev/pandas/pull/61041
    _FSSPEC_URL_PATTERN = re.compile(
        r"^[A-Za-z][A-Za-z0-9+\-+.]*(::[A-Za-z0-9+\-+.]+)*://"
    )

    def is_fsspec_url(url):
        return (
            isinstance(url, str)
            and bool(_FSSPEC_URL_PATTERN.match(url))
            and not url.startswith(("http://", "https://"))
        )

    pandas.io.common.is_fsspec_url = is_fsspec_url  # type: ignore
